######################################################################
#####
#####    PONGsim (Python Overlay Network Graphical Simulator)
#####    Module implementing the SIL-model 
#####    Version: 6.1
#####
######################################################################
#####    
#####    PONGsim (Python Overlay Network Graphical Simulator)
#####    
#####    Copyright (c) 2005-2007 Nicklas Beijar
#####    
#####    Permission is hereby granted, free of charge, to any person
#####    obtaining a copy of this software and associated documentation
#####    files (the "Software"), to deal in the Software without
#####    restriction, including without limitation the rights to use,
#####    copy, modify, merge, publish, distribute, sublicense, and/or
#####    sell copies of the Software, and to permit persons to whom the
#####    Software is furnished to do so, subject to the following
#####    conditions:
#####    
#####    The above copyright notice and this permission notice shall be
#####    included in all copies or substantial portions of the Software.
#####    
#####    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#####    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#####    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#####    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#####    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#####    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#####    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#####    OTHER DEALINGS IN THE SOFTWARE.
#####
######################################################################
#####
#####    Nicklas Beijar, nbeijar@netlab.tkk.fi
#####    Networking Laboratory, Helsinki University of Technology
##### 
######################################################################
#####
#####    To do:
#####    - (nothing)
#####
######################################################################

import pongsim
from Tkinter import *
import random, string

SIL_LINK_SEARCH 		= 0
SIL_LINK_INDEX 		= 1
SIL_LINK_SEARCH_TTL 	= 2
SIL_LINK_INDEX_TTL 	= 3
SIL_LINK_NODE 		= 4

SIL_MSG_SEARCH             = 2
SIL_MSG_INDEX                = 1
    
######################################################################
#####    S I L A P P L I C A T I O N
######################################################################
#####
#####    Framework for an application based on the Search/Index Link model.
#####    To be subclassed.
#####
######################################################################

class SILApplication(pongsim.Application):

    # May be subclassed. Subclass MUST call the parent class. 
    def init(self):
        self.seencache = []
        self.seencachesize = 200
        self.ctr_searchforwards = pongsim.Counter("searchforwards")
        self.links = []

    ##### HANDLERS FOR MESSAGES IMPLEMENTED BY SUBCLASS

    # An index update message is received. Returns the (possibly updated) message and the destinations to 
    # which the message should be forwarded.
    # To be subclassed. 
    def receiveindex(self, ttl, msg, destinations):
        return destinations
        
    # An index update message is received. Returns the (possibly updated) message and the destinations to 
    # which the message should be forwarded.
    # To be subclassed. 
    def receivesearch(self, ttl, msg, destinations):
        return (msg, destinations)

    # To be subclassed. 
    def receivegeneral(self, ttl, type, id, msg):
        return (msg, destinations)

    ##### UTILITY

    # NOT recommended to be subclassed. Returns True if message with given id 
    # has been forwarded by this node. Also remembers the current id. 
    # CAN be subclassed to skip checking. 
    def isnew(self, id):
        if id in self.seencache: return False
        self.seencache.append(id)
        if len(self.seencache) > self.seencachesize: del self.seencache[0]
        return True
        
    ##### SENDING, RECEIVING, FORWARDING MESSAGES
        
    # NOT to be subclassed
    def receive(self, packedmsg):
        (ttl, type, id, msg, prevnode) = packedmsg
        if type == SIL_MSG_INDEX: # index
            self.node.count("received-index-msg")
            if self.isnew(id):
                self.node.count("received-unique-index-msg")
                destinations = []
                for link in self.links:
                    if link[SIL_LINK_INDEX]:
                        if link[SIL_LINK_NODE] != prevnode and link[SIL_LINK_NODE] != self.node:
                            destinations.append(link[SIL_LINK_NODE])
                forwards = self.receiveindex(ttl, msg, destinations)
                for forward in forwards:
                    (msg, destinations) = forward
                    if destinations and ttl > 1:
                        self.node.count("forwarded-index-msg", len(destinations))
                        self.forwardindex(msg, destinations, ttl-1, id)
        elif type == SIL_MSG_SEARCH: # search
            self.node.count("received-search-msg")
            if self.isnew(id):
                self.node.count("received-unique-search-msg")
                destinations = []
                for link in self.links:
                    if link[SIL_LINK_SEARCH]:
                        if link[SIL_LINK_NODE] != prevnode and link[SIL_LINK_NODE] != self.node: 
                            destinations.append(link[SIL_LINK_NODE])
                firsthop = prevnode == self.node
                forwards = self.receivesearch(ttl, msg, destinations, firsthop)
                for forward in forwards:
                    (msg, destinations) = forward
                    if destinations and ttl > 1:
                        self.node.count("forwarded-search-msg", len(destinations))
                        self.forwardsearch(msg, destinations, ttl-1, id)
        else: # general
            self.node.count("received-general-msg")
            self.receivegeneral(ttl, type, id, msg)

    # NOT to be subclassed.
    # Called to send a search query on all outgoing search links. 
    # If no ttl is specified, the link-specific one is used. 
    # If no id is specified, a random one is generated. 
    def floodsearch(self, msg, ttl = None, id = None):
        if id == None:
            id = random.randint(-30000, 30000) # generate a random id if none is given
        self.node.count("sent-search-msg")
        self.receive((ttl+1, SIL_MSG_SEARCH, id, msg, self.node))
    
    # NOT to be subclassed.
    # Called to forward a search query on all outgoing search links. 
    # If no ttl is specified, the link-specific one is used. 
    def forwardsearch(self, msg, destinations, ttl = None, id = None):
        for dest in destinations:
            linkttl = ttl
            if linkttl == None: 
                for link in self.links:
                    if link[SIL_LINK_SEARCH]:
                        if link[SIL_LINK_NODE]==dest:
                            linkttl = link[SIL_LINK_INDEX_TTL]
            self.node.send((linkttl, SIL_MSG_SEARCH, id, msg, (self.node,)), dest)
    
    # NOT to be subclassed.
    # Called to send an index update on all outgoing index links. 
    # If no ttl is specified, the link-specific one is used. 
    # If no id is specified, a random one is generated. 
    def floodindex(self, msg, ttl = None, id = None):
        if id == None:
            id = random.randint(-30000, 30000) # generate a random id if none is given
        self.node.count("sent-index-msg")
        self.receive((ttl+1, SIL_MSG_INDEX, id, msg, self.node))

    # NOT to be subclassed.
    # Called to forward an index update on all outgoing index links. 
    # If no ttl is specified, the link-specific one is used. 
    # If no id is specified, a random one is generated. 
    def forwardindex(self, msg, destinations, ttl = None, id = None):    
        for dest in destinations:
            linkttl = ttl
            if linkttl == None: 
                for link in self.links:
                    if link[SIL_LINK_INDEX]:
                        if link[SIL_LINK_NODE]==dest:
                            linkttl = link[SIL_LINK_INDEX_TTL]
            self.node.send((linkttl, SIL_MSG_INDEX, id, msg, (self.node,)), dest)
    
    # NOT recommended to be subclassed.
    # Called to send a general message to a specified node (independently on links).
    # TTL and ID defaults to None, since these messages are not affected by flooding. They can be used for an application specific purpose. 
    # Type can be any integer, except for the reserved values (1 and 2), defaulting to 0. 
    def sendgeneral(self, msg, dest, type = 0, ttl = None, id = None):
        self.node.send((ttl, type, id, msg, (self.node,)), dest)

    ##### LINK MANAGEMENT

    # Adds a search or index link (or both) to node dest. 
    # For an index link: index = True, indexttl defined (default 255)
    # For a search link: search = True, searchttl defined (default 255)
    # If there already exists a link to the given node, that is updated. 
    def addlink(self, dest, index, search, indexttl = 255, searchttl = 255):
        assert index or search
        isanewlink = True
        for i in range(len(self.links)):
            if self.links[i][3] == dest:
                isanewlink = False
                if searchttl == None: searchttl = self.links[i][2]
                if indexttl == None: indexttl = sefl.links[i][3]
                self.links[i] = (self.links[i][0] or search, self.links[i][1] or index, searchttl, indexttl, dest)
        if isanewlink: self.links.append((search, index, searchttl, indexttl, dest))
    
    def removelink(self, dest, index, search):
        assert False # not implemented
    
    ##### USER INTERFACE
    
    def draw(self, canvas, x, y):
        for link in self.links:
            if link[0] and link[1]: # search and index
                fill = "#000"
            elif link[0]:               # search
                fill = "#F00"
            elif link[1]:               # index
                fill = "#00F"
            else:                       # no link
                fill = "#FFF"
            destx = link[4].getx()
            desty = link[4].gety()
            if not link[4].world:       # link to inexistent node
                destx = x + 40
                desty = y - 40
            canvas.create_line(x, y, link[4].getx(), link[4].gety(), fill=fill, width=1, arrow=LAST)


if __name__ == '__main__':
    print "This is a framework for p2p simulations, and is not runnable as such."
    print "To use it, create subclasses of the framework classes."
    