######################################################################
#####
#####    Simulator for search algorithms in unstructured peer-to-peer networks 
#####    Based on the PONGsim framework
#####    Version: 6.20
#####
######################################################################
#####
#####    Simulator for search algorithms in unstructured peer-to-peer networks 
#####    
#####    Copyright (c) 2005-2007 Nicklas Beijar
#####    
#####    Permission is hereby granted, free of charge, to any person
#####    obtaining a copy of this software and associated documentation
#####    files (the "Software"), to deal in the Software without
#####    restriction, including without limitation the rights to use,
#####    copy, modify, merge, publish, distribute, sublicense, and/or
#####    sell copies of the Software, and to permit persons to whom the
#####    Software is furnished to do so, subject to the following
#####    conditions:
#####    
#####    The above copyright notice and this permission notice shall be
#####    included in all copies or substantial portions of the Software.
#####    
#####    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#####    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#####    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#####    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#####    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#####    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#####    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#####    OTHER DEALINGS IN THE SOFTWARE.
#####
######################################################################
#####
#####    Nicklas Beijar, nbeijar@netlab.tkk.fi
#####    Networking Laboratory, Helsinki University of Technology
##### 
######################################################################
#####
#####    To do:
#####    - Everything
#####
######################################################################

from Tkinter import *
import random, string, math
import pongsim, pongsim_topologygenerator, pongsim_sil

MSG_CONNECT_REQ = 10
MSG_CONNECT_REP = 11
MSG_HELLO = 12
MSG_SEARCH_REP = 13

CONNECT_REQ_FIELD_NODE = 0
CONNECT_REQ_FIELD_CLUSTER = 1
CONNECT_REQ_FIELD_TTL = 2

CONNECT_REP_FIELD_NODE = 0
CONNECT_REP_FIELD_CLUSTER = 1

CLUSTER_STATUS_UNCONNECTED = 0
CLUSTER_STATUS_CONNECTING = 1
CLUSTER_STATUS_CONNECTED = 2

####################################################################
###   S I L A p p l i c a t i o n
####################################################################
class SILApplication(pongsim_sil.SILApplication):

    def init(self):
        """ Initializes the application.
        """
        pongsim_sil.SILApplication.init(self) # calling parent initiator is mandatory
        self.inboundnodes = {}
        self.clusterconnectstatus = {}
        self.clusterofaddress = {}
        self.connectedpeers = 0
        self.cluster = None
        self.localresources = []
        self.remoteresources = []
        self.neighborload = {}
        self.pendingsearches = {}
        self.pendingsearchtimes = {}
        scheduler.after(random.randint(0, sm.get("hellointerval")), self.hellointerval, None)
        scheduler.after(random.expovariate(1.0/sm.get("avgsearchinterval")), self.generatesearch, None) 
        scheduler.after(random.randint(0, sm.get("indexupdateinterval")), self.generateindexupdate, None)
        if sm.get("indexaddinterval", None):
            scheduler.after(random.expovariate(1.0/sm.get("indexaddinterval")), self.addresource, None)
        
    ############ USER INTERFACE

    def getdisplayinfo(self):
        """ Returns the text written beside the node. 
        """
        return str(self.node.address)+"/"+str(self.cluster)

    def getdetailedinfo(self):
        """Returns the the node's description that is written in a separate window when the node is examined. 
        """
        t = "Cluster: " + str(self.cluster) + "\n"
        t = t + "Counters:\n"
        for k in self.node.counters.keys():
            t = t + "   " + k + ": " + str(self.node.counters[k]) + "\n"
        t = t + "Neighborload: "
        for k in self.neighborload.keys():
            t = t + str(k.address) + ": " + str(self.neighborload[k]) + "   "
        t = t + "\n"
        t = t + "Search links:\n"
        for link in self.links:
            if link[pongsim_sil.SIL_LINK_SEARCH]:
                t = t + "   " + str(link[pongsim_sil.SIL_LINK_NODE].address) + "/" + str(link[pongsim_sil.SIL_LINK_NODE].application.cluster) + "\n"
        t = t + "Index links:\n"
        for link in self.links:
            if link[pongsim_sil.SIL_LINK_INDEX]:
                t = t + "   " + str(link[pongsim_sil.SIL_LINK_NODE].address) + "/" + str(link[pongsim_sil.SIL_LINK_NODE].application.cluster) + "\n"
        t = t + "Local resources:\n"
        t2 = "   "
        for resource in self.localresources:
            t2 = t2 + str(resource) + "  "
            if len(t2) > 30:
                t = t + t2
                t2 = "\n   "
        t = t + t2+"\n"    
        t = t + "Remote resources:\n"
        t2 = "   "
        for resource in self.remoteresources:
            t2 = t2 + str(resource[0]) + "@" + str(resource[1]) + "  "
            if len(t2) > 30:
                t = t + t2
                t2 = "\n   "
        t = t + t2+"\n"    
        return t

    ############ GENERAL SUPPORT FUNCTIONS

    def outofbanddata(self, key, value):
        """ Sets the data that is agreed without any protocol (operator agreements) or manually configured.
        
        Corresponds to the user/administrator configuring the application. 
        """
        if key == "inboundnodes":
            self.inboundnodes = value # inbound node of every operator in the network
            for cluster in self.inboundnodes.keys():
                self.clusterconnectstatus[cluster] = CLUSTER_STATUS_UNCONNECTED
        elif key=="resource":
            self.localresources.append(value)
        elif key=="cluster":
            self.cluster = value
        else:
            print "Unknown key for outofbanddata"

    def receivegeneral(self, ttl, type, id, msg):
        """ Handles reception of a general message and calls the corresponding handler. 
        
        Called by the SIL framework.
        """
        if not self.node.world: return # scheduled functions need to check the node's existence
        if type == MSG_CONNECT_REQ:
            self.received_connectreq(msg)
        elif type == MSG_CONNECT_REP:
            self.received_connectrep(msg)
        elif type == MSG_HELLO:
            self.received_hello(msg)
        elif type == MSG_SEARCH_REP:
            self.received_searchrep(msg)
        else:
            print "ERROR: Unknown general message:",type

    def hellointerval(self, dummy=None):
        """ Performs the functions that are done periodically at each hello interval.
        """
        if not self.node.world: return # scheduled functions need to check the node's existence
            
        # send hello messages
        self.generatehello()
            
        # check if some cluster is unconnected
        for cluster in self.inboundnodes.keys():
            if cluster != self.cluster:
                if self.clusterconnectstatus[cluster] == CLUSTER_STATUS_UNCONNECTED:
                    scheduler.after(random.randint(1,120), self.connectsearchlink, (self.inboundnodes[cluster], cluster))
                
        # schedule next hello
        scheduler.after(sm.get("hellointerval"), self.hellointerval, None)

    ############ HELLO PROTOCOL

    def generatehello(self):
        """ Sends a hello message on all outgoing index links.
        """
        if not self.node.world: return # scheduled functions need to check the node's existence
        for link in self.links:
            if link[pongsim_sil.SIL_LINK_INDEX]:
                self.sendgeneral((self.node, self.cluster, self.connectedpeers), link[pongsim_sil.SIL_LINK_NODE], MSG_HELLO)

    def received_hello(self, msg):
        """ Handles a received hello message. 
        
        Store the load of this neighbor. 
        """
        if msg[1] != self.cluster:
            print "WARNING: Received hello from another cluster - ignoring"
        else:
            self.neighborload[msg[0]] = msg[2]

    ############ SEARCH LINK CONNECTION PROTOCOL

    def connectsearchlink(self, params):
        """ Sends a connect request to another cluster for connecting a search link. 
        """
        if not self.node.world: return # scheduled functions need to check the node's existence
        (node, cluster) = params
        # print "Node",self.node.address,"in cluster",self.cluster,"sends CONNECT REQ to node",node.address,"in cluster",cluster
        if self.clusterconnectstatus[cluster] != CLUSTER_STATUS_UNCONNECTED:
            pass # this happens and a lot (cluster connect delay > hello check interval)
            #print "WARNING: Would have tried to connect to already connected cluster"
        else:
            self.clusterconnectstatus[cluster] = CLUSTER_STATUS_CONNECTING        
            self.sendgeneral((self.node, self.cluster, 5), node, MSG_CONNECT_REQ)

    def received_connectreq(self, msg):
        """ Handles a received connect request. 
        
        Forward to a neighbor with a lower load if possible. Otherwise answer to the request with a connect reply. 
        """
        if msg[CONNECT_REQ_FIELD_CLUSTER] == self.cluster:
            print "WARNING: Node",self.node.address,"received connection attempt within the same cluster:",self.cluster,"- ignoring"
        else:
            lowestload = self.connectedpeers
            lowestpeer = self.node
            for k in self.neighborload:
                if self.neighborload[k] < lowestload: 
                    lowestload = self.neighborload[k]
                    lowestpeer = k
            if lowestpeer == self.node or msg[CONNECT_REQ_FIELD_TTL] < 1:
                self.sendgeneral((self.node,self.cluster), msg[CONNECT_REQ_FIELD_NODE], MSG_CONNECT_REP)
                self.connectedpeers += 1
            else:
                self.sendgeneral((msg[CONNECT_REQ_FIELD_NODE], msg[CONNECT_REQ_FIELD_CLUSTER], msg[CONNECT_REQ_FIELD_TTL]-1), lowestpeer, MSG_CONNECT_REQ)

    def received_connectrep(self, msg):
        """ Handles a received connect reply and adds the new search link. 
        """
        if self.clusterconnectstatus[msg[CONNECT_REP_FIELD_CLUSTER]] == CLUSTER_STATUS_CONNECTED:
            print "WARNING: Received connect reply from already connected cluster"
        else:
            # print "Search link",self.node.address,"/",self.cluster,"->",msg[0].address,"/",msg[1]
            self.clusterconnectstatus[msg[CONNECT_REP_FIELD_CLUSTER]] = CLUSTER_STATUS_CONNECTED
            self.clusterofaddress[msg[CONNECT_REP_FIELD_NODE]] = msg[CONNECT_REP_FIELD_CLUSTER]
            self.addlink(msg[CONNECT_REP_FIELD_NODE], False, True, None, 1)
            self.node.count_override("searchlinks")

    ############ DYNAMIC REOURCES
    
    def addresource(self, dummy=None):
        """Dynamically add a resource to the node. 
        
        Adds a random resource, id defined by the scenario-parameters dynamicresourcefrom and dynamicresourceto.
        If dynamicresourceto is undefined, nrofresources is used as upper limit. 
        Sends an index update to inform other nodes about the new resource. 
        Schedules itself again after interval defined in the scenario-parameter indexaddinterval.
        Schedules removal of the added resource after interval defined in the scenario-parameter indexremoveinterval
        (or in indexaddinterval if indexremoveinterval is undefined).
        """
        if not self.node.world: return # scheduled functions need to check the node's existence
        newresource = random.randint(sm.get("dynamicresourcefrom", 0), sm.get("dynamicresourceto", sm.get("nrofresources")))
        self.localresources.append(newresource)
        msg = (self.node, (newresource,), None)
        self.floodindex(msg, ttl = 256)
        scheduler.after(random.expovariate(1.0/sm.get("indexremoveinterval", sm.get("indexaddinterval"))), self.removeresource, newresource)
        scheduler.after(random.expovariate(1.0/sm.get("indexaddinterval")), self.addresource, None)

    def removeresource(self, resource):
        """Dynamically remove a resource from the node. 
        
        The removed resource is given in the resource argument. 
        Sends an index update to inform other nodes about the removed resource. 
        """
        if not self.node.world: return # scheduled functions need to check the node's existence
        if resource not in self.localresources: return
        for i in range(len(self.localresources)):
            if self.localresources[i] == resource:
                break
        del self.localresources[i]
        msg = (self.node, None, (resource,))
        self.floodindex(msg, ttl = 256)
        
    ############ INDEX PROTOCOL
    
    def generateindexupdate(self, dummy=None):
        """Generates a periodical index update. 
        
        Schedules itself after time defined in scenario-parameter indexupdateinterval.
        """
        if not self.node.world: return # scheduled functions need to check the node's existence
        msg = (self.node.address, self.localresources, None)
        self.floodindex(msg, ttl = 256)
        scheduler.after(sm.get("indexupdateinterval"), self.generateindexupdate, None)

    def receiveindex(self, ttl, msg, destinations):
        """Handles reception of index update.
        
        Called by framework on reception of index update. 
        Adds and deletes remote resources specified in the index update. 
        Lets forwarding the index update continue.
        """
        (owner, additions, deletes) = msg
        if additions:
            for resource in additions:
                if resource not in self.localresources:
                    found = None
                    for i in range(len(self.remoteresources)):
                        if self.remoteresources[i][0] == resource and self.remoteresources[i][1] == owner:
                            found = i
                    if not found: self.remoteresources.append((resource, owner))
        if deletes:
            for resource in deletes:
                found = None
                for i in range(len(self.remoteresources)):
                    if self.remoteresources[i][0] == resource and self.remoteresources[i][1] == owner:
                        found = i
                if found: del self.remoteresources[found]
        return [(msg, destinations), ]
        
    ############ SEARCH PROTOCOL

    def generatesearch(self, dummy=None):
        """ Generates a search request, which is flooded on the search links. 
        
       Scheduled calling, with rescheduling. 
        """
        if not self.node.world: return # scheduled functions need to check the node's existence
        resource = random.randint(0, sm.get("nrofresources")-1)
        searchid = random.randint(0, 15000000)
        self.pendingsearches[searchid] = []
        self.pendingsearchtimes[searchid] = None
        searcher = self.node
        self.node.count("generated-searches")
        clusterpath = [] ## [self.cluster,]
        clusterstack = []
        sendtime = scheduler.time()
        msg = (searcher, searchid, resource, clusterpath, clusterstack, sendtime)
        self.floodsearch(msg, ttl = int(sm.get("searchttl")))
        scheduler.after(sm.get("searchtimeout"), self.searchtimeout, searchid) # schedule timeout for checking results
        scheduler.after(random.expovariate(1.0/sm.get("avgsearchinterval")), self.generatesearch, None)  # schedule next search

    def receivesearch(self, ttl, msg, destinations, firsthop = False):
        """ Handles received search requests.
        
        Called by the SIL framework. 
        Determines the nodes to which the message is forwarded from the candidate destinations based on the search strategy. 
        Maintains clusterpath and clusterstack if configured so.         
        """
        # Unpack message.
        (searcher, searchid, resource, clusterpath, clusterstack, sendtime) = msg
        #print "A",msg,destinations
        ##print searchid,": in  ",self.node.address,"/",self.cluster,"S",clusterstack, "P",clusterpath, "D",destinations
        
        # Make new instances
        newdestinations = []
        for dest in destinations: newdestinations.append(dest)
        destinations = newdestinations
        newclusterpath = []
        for cluster in clusterpath: newclusterpath.append(cluster)
        clusterpath = newclusterpath
        newclusterstack = []
        for cluster in clusterstack: newclusterstack.append(cluster)
        clusterstack = newclusterstack
                
        # Check if the result was found, and reply with answer. Stop the forwarding if 
        # "forward-on-match" is False.
        found = None
        for res in self.remoteresources:
            if res[0] == resource: found = res[1]
        if resource in self.localresources:
            found = self.node.address
        if (found != None): 
            self.node.count("found-match")
            self.replysearch(searcher, searchid, sendtime, found)
            if not sm.get("forward-on-match"): return [(msg, []), ]
                
        # If "trackcluster-precheck" is True, then drop all candidate neighbors whose cluster are in 
        # the clusterpath already or in the current cluster.
        if sm.get("trackcluster-precheck"):
            newdestinations = []
            for dest in destinations:
                if self.clusterofaddress.has_key(dest):
                    if (self.clusterofaddress[dest] not in clusterpath) and (self.clusterofaddress[dest] != self.cluster):
                        newdestinations.append(dest)
                else:
                    print "WARNING: Trackcluster-precheck found neighbor belonging to unknown cluster"
                    newdestinations.append(dest)
            destinations = newdestinations
        
        # Forward according to the "searchstrategy":
        # randomwalk:
        #       Select a random neighbor from the candidate neihbors.
        # flood:
        #       Forward to all candidate neighbors. 
        # stackrandomwalk:
        #       Add to the stack all candidate neighbors whose cluster is not already in the stack.
        #       Then take one node from the stack and forward there. 
        #       NOTE: Stackrandomwalk requires trackcluster-precheck to be True.
        if sm.get("searchstrategy") == "randomwalk":
            if len(destinations) > 0: 
                destinations = [destinations[random.randint(0, len(destinations)-1)], ]
        elif sm.get("searchstrategy") == "flood":
            pass
        elif sm.get("searchstrategy") == "stackrandomwalk":
            for dest in destinations:
                found = False
                for (stackcluster, stackaddress) in clusterstack:
                    if stackcluster == self.clusterofaddress[dest]: found = True
                if not found:
                    clusterstack.append((self.clusterofaddress[dest], dest))
            destinations = []
            if len(clusterstack) > 0:
                (stackcluster, stackaddress) = clusterstack.pop(-1)
                destinations.append(stackaddress)
        elif sm.get("searchstrategy") == "hsbrw":
            for dest in destinations:
                found = False
                for (stackcluster, stackaddress) in clusterstack:
                    if stackcluster == self.clusterofaddress[dest]: found = True
                if not found:
                    clusterstack.append((self.clusterofaddress[dest], dest))
            destinations = []
            numforwards = int(sm.get("pops", default=1))
            for i in range(numforwards):
                if len(clusterstack) > 0:
                    (stackcluster, stackaddress) = clusterstack.pop(-1)
                    destinations.append(stackaddress)
        else:
            print "ERROR: Unknown search strategy:",sm.get("searchstrategy")

        # If "trackcluster" is True, then drop the whole message if it has been in the same cluster
        # before - otherwise add the cluster to the path and continue forwarding. 
        if sm.get("trackcluster"):
            if self.cluster in clusterpath:
                destinations = []
                self.node.count("track-cluster-block")
            else:
                self.node.count("track-cluster-forward")
                clusterpath.append(self.cluster)
                
        # Construct new forwarded message and continue forwarding. 
        forward = []
        if sm.get("searchstrategy") == "hsbrw":
            for i in range(len(destinations)):
                thisclusterstack = []
                j = i
                while j < len(clusterstack):
                    thisclusterstack.append(clusterstack[j])
                    j += len(destinations)
                #print i,clusterstack,thisclusterstack
                forward.append(((searcher, searchid, resource, clusterpath, thisclusterstack, sendtime), [destinations[i], ]))
        else:
            forward.append(((searcher, searchid, resource, clusterpath, clusterstack, sendtime), destinations))
        return forward

    def replysearch(self, searcher, searchid, sendtime, where):
        """Send a search reply. 
        
        Called by receivesearch on a found match. 
        """
        self.sendgeneral((self.node, self.cluster, searchid, sendtime, where), searcher, MSG_SEARCH_REP)
        self.node.count("sent-searchrep")

    def received_searchrep(self, msg):
        """Handles a received search reply. 
        
        Count the answers. 
        """
        (found_node, found_cluster, searchid, sendtime, where) = msg
        self.node.count("received-searchrep")
        if self.pendingsearches.has_key(searchid):
            if found_node not in self.pendingsearches[searchid]:
                self.node.count("received-searchrep-unique")
                self.pendingsearchtimes[searchid] = scheduler.time()-sendtime
            self.pendingsearches[searchid].append(where)
            self.node.world.simulator.reporter.multivalue("search-delay", scheduler.time()-sendtime)
            if len(self.pendingsearches[searchid]) == 1:
                self.node.world.simulator.reporter.multivalue("search-delay-first", scheduler.time()-sendtime)
        else:
            print "WARNING: Result arrived for nonexistent or timed out search. It's OK, but statistics not collected. Try increasing the timeout."

    def searchtimeout(self, searchid):
        """Examines the search results collected during the search, and collects statistics about the search.
        
        Called when a given time has expired since the search was initiated. A search request timed out with either results or without results. 
        """
        if not self.node.world: return # scheduled functions need to check the node's existence
        if (not self.pendingsearches.has_key(searchid)) or (not self.pendingsearchtimes.has_key(searchid)):
            print "WARNING: Search timeout for nonexistent search"
            return
        if len(self.pendingsearches[searchid]) > 0:
            self.node.count("search-found")
            self.node.world.simulator.reporter.multivalue("search-responses", len(self.pendingsearches[searchid]))
            self.node.world.simulator.reporter.multivalue("search-delay-last", self.pendingsearchtimes[searchid])
            unique = []
            for node in self.pendingsearches[searchid]:
                if node not in unique: unique.append(node)
            self.node.world.simulator.reporter.multivalue("search-found-copies", len(unique))
        else:
            self.node.count("search-not-found")
        del self.pendingsearches[searchid]
        del self.pendingsearchtimes[searchid] 


####################################################################
###   S I L S i m u l a t i o n
####################################################################
class SILSimulation(pongsim.Simulation):
 
    def init(self):
        """Initiates the simulation.
        """
        random.seed(self.scenariomanager.get("randomseed")+self.scenariomanager.getrunnumber())
        self.datacollectiondelay = 1000
        if self.scenariomanager.get("clusterpreset", None) != None:
            preset = self.scenariomanager.get("clusterpreset", None)
            if preset == 1:
                clustersizes = string.splitfields("5;5;5;5;5", ";")
            elif preset == 2:
                clustersizes = string.splitfields("4;4;5;6;6", ";")
            elif preset == 3:
                clustersizes = string.splitfields("3;4;5;6;7", ";")
            elif preset == 4:
                clustersizes = string.splitfields("2;3;5;7;8", ";")
            elif preset == 5:
                clustersizes = string.splitfields("1;2;5;8;9", ";")
            elif preset == 6:
                clustersizes = string.splitfields("1;1;5;9;9", ";")
        elif self.scenariomanager.get("clustersizes", None) == None:
            numberofclusters = self.scenariomanager.get("clusters", 1)
            nodesincluster= self.scenariomanager.get("nodesincluster", 1)
            s = ""
            for i in range(int(numberofclusters)):
                s += str(int(nodesincluster)) + ";"
            clustersizes = string.splitfields(s[:-1], ";")
        else:
            clustersizes = string.splitfields(self.scenariomanager.get("clustersizes"), ";")
        numberofclusters = len(clustersizes)
        
        self.allnodes = []
        clusternodes = []
        inboundnodes = {}
        for c in range(numberofclusters):
            clusterx = self.world.xsize/2.0*0.7*math.cos(float(c)/float(numberofclusters)*2.0*math.pi) + self.world.xsize/2.0
            clustery = self.world.ysize/2.0*0.7*math.sin(float(c)/float(numberofclusters)*2.0*math.pi) + self.world.ysize/2.0
                
            nodes_in_cluster= []
            #nodesincluster = random.randint(self.scenariomanager.get("minnodesincluster"), self.scenariomanager.get("maxnodesincluster"))
            nodesincluster = int(clustersizes[c])
            # (clusterx, clustery) = self.world.randomcoordinate()
            for i in range(nodesincluster):
                #(nodex, nodey) = self.world.randomcoordinateclose(clusterx, clustery, 0.25)
                nodex = self.world.xsize*0.1*math.cos(float(i)/float(nodesincluster)*2.0*math.pi) + clusterx
                nodey = self.world.ysize*0.1*math.sin(float(i)/float(nodesincluster)*2.0*math.pi) + clustery
                newnode = pongsim.Node(self.world, color = "#F00", x=nodex, y=nodey)
                newapp = SILApplication(self.world, newnode)
                newnode.application.outofbanddata("cluster", c)
                nodes_in_cluster.append(newnode)
                self.allnodes.append(newnode)
            clusternodes.append(nodes_in_cluster)
            
            for i in range(len(nodes_in_cluster)):
                peernode = i+1
                if peernode==len(nodes_in_cluster): peernode=0
                # print "Index link",nodes_in_cluster[i].address,"/",nodes_in_cluster[i].application.cluster,"->",nodes_in_cluster[peernode].address,"/",nodes_in_cluster[peernode].application.cluster
                nodes_in_cluster[i].application.addlink(nodes_in_cluster[peernode], True, False, 100, 100)
                nodes_in_cluster[i].count_override("indexlinks")
        
        #for n in inboundnodes.keys():
        #    print n, inboundnodes[n].address, inboundnodes[n].application.cluster
        self.generate_cluster_topology(clusternodes)
        self.generate_resources(self.allnodes)
        
        self.collectdata = False
        scheduler.after(self.datacollectiondelay, self.startdatacollection, None)
        
    def generate_resources(self, allnodes):
        """Generates the shared resources and allocates them to the nodes.
        """
        for resource in range(int(self.scenariomanager.get("nrofresources", 100))):
            if self.scenariomanager.get("mincopies", None) != None and self.scenariomanager.get("maxcopies", None) != None:
                copies = random.randint(self.scenariomanager.get("mincopies"), self.scenariomanager.get("maxcopies"))
            else:
                copies = int(self.scenariomanager.get("copies", 1))
            given = []
            for i in range(copies):
                while 1:
                    node = random.randint(0, len(allnodes)-1)
                    if node not in given: break
                given.append(node)
                allnodes[node].application.outofbanddata("resource", resource)
        
    def generate_cluster_topology(self, clusternodes):
        """Generates the topology connecting clusters together according to the chosen scenario.
        """
        while 1:
            clustertopology = pongsim_topologygenerator.Topology(len(clusternodes))
            if sm.get("searchtopology") == "powerlaw":
                clustertopology.powerlaw_BA_degree(sm.get("avgdegree"))
                self.reporter.multivalue("real-avg-degree", clustertopology.avgdegree())
            elif sm.get("searchtopology") == "fullyconnected":
                clustertopology.fullyconnected()
            else:
                print "Unknown search topology"
            if clustertopology.isconnected(): 
                break
            else:
                print "Unconnected topology -> Regenerating"
        for c in range(len(clusternodes)):
            cluster_inbound = {}
            for n in clustertopology.neighbors(c):
                cluster_inbound[n] = clusternodes[n][0]      # add first node of cluster as inbound node
            for node in clusternodes[c]:
                node.application.outofbanddata("inboundnodes", cluster_inbound)

    def close(self):
        """Closes the simulator and collects the statistics.
        """
        sim_time = self.scheduler.time()
        for node in self.allnodes:
            for counter in node.counterlist():
                self.reporter.multivalue(counter, node.counter(counter) / (sim_time-self.datacollectiondelay))
            #self.reporter.multivalue("received-search-msg", node.counter("received-search-msg") / sim_time)
            #self.reporter.multivalue("sent-search-msg", node.counter("sent-search-msg") / sim_time)
            #self.reporter.multivalue("received-unique-search-msg", node.counter("received-unique-search-msg") / sim_time)

    def startdatacollection(self, dummy):
        """Allows collecting the statistics.
        """
        self.collectdata = True

    def stopcriteria(self):
        """Defines whether the simulation should end.
        """
        if self.scheduler.time() > self.scenariomanager.get("simtime", 5000): return True
        return False

####################################################################
###   M a i n   P r o g r a m
####################################################################


if __name__ == '__main__':
    global scheduler, sm, simulator
    scheduler = pongsim.Scheduler()
    sm = pongsim.ScenarioManager()
    sm.loadall(".\\scenarios\\")
    master = Tk()
    simulator = SILSimulation(scheduler, master, sm)
    simulator.setworld(pongsim.World(1000, 800))
    simulator.setoutputdir(".\\output\\")
    master.mainloop()
    

